import {throttle} from './hide-incognito-mode-util.mjs';

const showLog = false;
let os = '';
function log() {
    showLog && console.log.apply(console, arguments);
}

const state = {
    aggressiveModeEnabled: false
};

// note this gets called on extension install + update, but also on chrome update.
function recordInstall(currentAppVersion, currentTime) {

    // They may have existing values. So, we request them.
    chrome.storage.sync.get({
        "appVersionHistory": [],
        "initialAppInstallTime": currentTime
    }, savedVals => {

        // Add the current version to the user's history, if not already there.
        const versions = savedVals.appVersionHistory;
        if (!versions.includes(currentAppVersion)) {
            versions.push(currentAppVersion);
        }

        // Now we write the values back, likely setting them for the first time.
        chrome.storage.sync.set({
            "appVersionHistory": versions,
            "initialAppInstallTime": savedVals.initialAppInstallTime,
            "mostRecentUpdateTime": currentTime,
        }, updateUninstallUrl);

    });
}

// max url length is 255 chars, and we actually hit this limit sometimes.
// So, we'll try to build the longest url we can, giving preference for args at the front of the array.
function buildUrlRespectingMaxLength(queryArgs) {
    let url = 'http://rehfeld.us/browser-extensions/uninstall-survey.php?';

    let argSep = '';
    for (const arg of queryArgs) {
        if (url.length + argSep.length + arg.length < 255) {
            url += argSep + arg;
            // Trick to prevent adding & on first loop iteration.
            argSep = '&';
        }
    }

    return url;
}

function updateUninstallUrl() {
    chrome.storage.sync.get(null, syncVals => {
    // chrome.storage.sync.get(["appVersionHistory", "initialAppInstallTime", "mostRecentUpdateTime"], syncVals => {
        chrome.storage.local.get({usageStats: {}}, localVals => {
            const manifest = chrome.runtime.getManifest();

            // Add the current version to the user's history, if not already there.
            const versions = (syncVals.appVersionHistory || []);
            if (!versions.includes(manifest.version)) {
                versions.push(manifest.version);
            }
            const sortedVersions = versions.sort();

            // To save space, we only send min and max versions.
            let versionRange = sortedVersions[0];
            if (sortedVersions.length > 1) {
                versionRange +=  "," + sortedVersions[sortedVersions.length - 1];
            }

            // chrome imposes a 255 char url limit, so we need to use compact query var names.
            const queryArgs2 = {
                app: manifest.extNameCode,
                it: syncVals.initialAppInstallTime,
                ut: syncVals.mostRecentUpdateTime,
                vh: versionRange,
                // os might not yet have a value, but oh well.
                os: os,
            };

            const extra = {
                am: syncVals.aggressiveModeEnabled ? 1 : 0,
                ama: syncVals.aggressiveModeAttemptedAtLeastOnce ? 1 : 0,
                id: syncVals.incognitoDetected ? 1 : 0,
            };

            const map = Object.assign({}, queryArgs2, extra);
            const urlKeyValPairs = Object.keys(map).map(key => {
                return encodeURIComponent(key) + '=' + encodeURIComponent(map[key]);
            });

            chrome.runtime.setUninstallURL(buildUrlRespectingMaxLength(urlKeyValPairs));
        });
    });
}

chrome.runtime.onInstalled.addListener(details => {
    log("onInstall", details);
    const manifest = chrome.runtime.getManifest();
    recordInstall(manifest.version, unixTimestamp());
});

// Record the os into a var.
chrome.runtime.getPlatformInfo(function(info) {
    os = info.os;
});

chrome.storage.onChanged.addListener(updateUninstallUrl);

function unixTimestamp() {
    return Math.floor(Date.now() / 1000);
}

// Open the setup / install / configuration instructions page when they install.
chrome.runtime.onInstalled.addListener(function (details) {
    log('onInstall', details);
    if (details.reason === chrome.runtime.OnInstalledReason.INSTALL) {
        // Note, the local link will close the instant the user sets "allow in incognito" or reloads the ext.
        // So, we link to the web to prevent the tab from closing. We just hope the website is up 24/7.
        const url = chrome.runtime.getURL('ext-web/setup.html');
        // const url = 'http://rehfeld.us/browser-extensions/hide-incognito-mode/setup.html';
        chrome.tabs.create({url: url}, function (tab) {
            log(`New tab launched for ${url}`);
        });
    }

    // Updates seem to trigger when the user changes the "allow in incognito" config, which is weird.
    // Anyway, we only want to show the user the update page when their version changes, and only if they update to/from a specific version, so we check for this.
    // if (details.reason === chrome.runtime.OnInstalledReason.UPDATE && chrome.runtime.getManifest().version === "1.8" && details.previousVersion === "1.7") {
    //     const url = chrome.runtime.getURL('ext-web/updated.html');
    //     // const url = 'http://rehfeld.us/browser-extensions/hide-incognito-mode/updated.html';
    //     chrome.tabs.create({url: url}, function (tab) {
    //         log(`New tab launched for ${url}`);
    //     });
    // }
});

let num = 0;
function animateBrowserIcon() {
    num++;

    // chrome.browserAction.setBadgeText({text: "Ok"});
    // chrome.browserAction.setBadgeText({text: "Clear" + num});
    chrome.browserAction.setIcon({path: "icons/bandit_mask_white_16x16.png"});

    setTimeout(() => {
    //     chrome.browserAction.setBadgeText({text: ""});
        chrome.browserAction.setIcon({path: "icons/bandit_mask_16x16.png"});
    }, 1000);

    log('cleared private data in bg script');
}

// const throttledClearData = throttle(clearBrowsingData, 3000, {trailing: true});
const throttledClearData = throttle(clearBrowsingData, 3000, {leading: true});
chrome.runtime.onMessage.addListener(function(request, sender, sendResponse) {
    log(request);
    if (request.stateUpdateMessage) {
        updateState(request);
        sendResponse({farewell: "goodbye"});
    }

    if (request.clearPrivateData) {
        if (request.callFromContentScript) {
            // Due to iframes, the extension will make many calls to clear private data per page load. Clearing cookies while the
            // page is mid load causes problems, so we try to avoid it. This isn't a very good way to do this though, but it kinda works, for now.
            throttledClearData(animateBrowserIcon);
        } else {
            // A call from a hotkey - we wont throttle it, ensuring it feels and responds snappy.
            clearBrowsingData(animateBrowserIcon);
        }
    }
});

chrome.commands.onCommand.addListener(function(command, tab) {
    log('command:', command, tab);
    if (command !== 'wipe-data') {
        return;
    }
    // This runs in background page.
    // This callback can get called without activeTab, but w/o that it cant access the tab.url nor other stuff.
    // if (!tab.incognito) {
    if (tab.incognito) {
        log('calling clearBrowsingData ');
        clearBrowsingData(function removeSuccessCallback() {
            log('clearBrowsingData removeSuccessCallback ');
        });
    }
});

chrome.storage.onChanged.addListener(function(changes, namespace) {
    getSavedOptions().then(updateState);
});

function getSavedOptions() {
    const aggressiveModeEnabledDefaultValue = false;
    return new Promise(resolve => {
        chrome.storage.sync.get({
            "aggressiveModeEnabled": aggressiveModeEnabledDefaultValue,
        }, resolve);
    });
}

function updateState(newState) {
    state.aggressiveModeEnabled = newState.aggressiveModeEnabled;
}

function clearBrowsingData(removeSuccessCallback) {
    const millisecondsPerWeek = 1000 * 60 * 60 * 24 * 7;
    const timeAgo = (new Date()).getTime() - (millisecondsPerWeek * 52 * 10);
    const removalOptions = {
        "since": timeAgo,
        "originTypes": {
            "unprotectedWeb": true
        }
    };

    const dataTypeSet = {
        "appcache": true,
        "cache": true,
        "cacheStorage": true,
        "cookies": true,
        // "downloads": true, // clears the list, not the files. no purposes imo.
        "fileSystems": true,
        "formData": true,
        // "history": true, // I think this prevents the back button. not good.
        "indexedDB": true,
        "localStorage": true,
        "serverBoundCertificates": true,
        "pluginData": true, // Deprecated since Chrome 88. Support for Flash has been removed. This function has no effect.
        "serviceWorkers": true,
        "webSQL": true
    };

    chrome.browsingData.remove(removalOptions, dataTypeSet, removeSuccessCallback);
}